"""
    This module implements the schemes of L-BFGS initial matrix tested in

    (GL)
    Some numerical experiments with variable-storage quasi-Newton algorithms,
    1989,
    Gilbert, J.C. & Lemarechal, C. Mathematical Programming (1989) 45: 407.

    doi:10.1007/BF01589113

    https://link.springer.com/article/10.1007/BF01589113

    Interesting quote from Gilbert and Lemarechal:

        Another point is that most QN-like methods modify the approximation of the
        Hessian by low rank corrections. As a result, w h e n / 4 , - Hk is a high rank matrix,
        a few QN updates have little effect on the improvement of Hk. On the other hand,
        scalings, which are full rank corrections, have a determining importance on the
        performance. In other words, least-change updates may be too shy in the present
        context and should perhaps be replaced by high rank stable updates.


    GL recommends using M1QN3.B2, which scales the diagonal D before the update.
    The scaling requires D to satisty the quasi-Cauchy condition, improving convergence.

        M1QN3_B2 = LBFGS(diag_update=pre_scaled_direct_bfgs, rescale_diag=False)

    In another reference, (VAF)  Limited-Memory BFGS Diagonal Preconditioners by
    F. VEERSE D. AUROUX and M. FISHER, Optimization and Engineering, 1, 323-339, 2000

    The authors showed scaling after D-update improves the convergence even more.

        VAFFast = LBFGS(diag_update=post_scaled_direct_bfgs, rescale_diag=False)

    VAF recommended a "new approach", where they do not scale D during D-update, but only
    scale D before L-BFGS update. The "new approach" was supposed to improve hessian
    approximation without suffering convergence rate.

        VAFGoodHessian = LBFGS(diag_update=direct_bfgs, rescale_diag=True)

    I only find slower convergence in our linear + smoothing and LPT-nonlinear cases.
    Never cheched the hessian.

    One may tempt to use,

        M1QN3_B2_RESCALED = LBFGS(diag_update=pre_scaled_direct_bfgs, rescale_diag=True)

    but it is as slow as VAFGoodHessian, and due the prescaling it doesn't give good hessian.

    We shall do these tests more rigourously and write some notes.

    We use the canonical scalar D-update as *default*. It seems to be more reliable than
    other methods for our cases, where the problem is quasi-linear, with complicated off
    diagonal structures. Perhaps in this case the low rank update is safer than the full
    rank update.

    Yu Feng
"""

from abopt.base import Optimizer
from abopt.base import Proposal

from abopt.linesearch import backtrace
from abopt.linesearch import simpleregulator

def scalar(vs, hessian):
    """ M1QN2.A in GL.  EQ 4.1;
        This is L-BFGS when people refers to it.
        wikipedia version of L-BFGS

    """

    assert len(hessian.S) > 0

    s = hessian.S[-1]
    y = hessian.Y[-1]
    ys = hessian.YS[-1]
    yy = hessian.YY[-1]

    D1 = ys / yy
    return D1

def inverse_bfgs(vs, hessian):
    """ 
        M1QN3.A in GL Equation 4.6.
        This is bad.  D grows rapidly over iterations.

    """
    assert len(hessian.S) > 0

    D0 = hessian.D
    s = hessian.S[-1]
    y = hessian.Y[-1]
    ys = hessian.YS[-1]
    yy = hessian.YY[-1]

    dot = vs.dot
    mul = vs.mul
    pow = vs.pow
    addmul = vs.addmul

    Dy = mul(y, D0)
    Dyy = dot(Dy, y)

    a3 = mul(mul(y, D0), s)

    D1 = addmul(D0, pow(s, 2), (1 / ys  + Dyy / ys ** 2))
    D1 = addmul(D1, mul(mul(y, D0), s), -2.0 / ys)

    return D1

def direct_bfgs(vs, hessian, pre_scaled=False, post_scaled=False):
    """ 
        M1QN3.B, M1QN3.B2, GL Equation 4.7, 4.9
        and VAF post update scaling.
    """
    assert len(hessian.S) > 0

    D0 = hessian.D

    s = hessian.S[-1]
    y = hessian.Y[-1]
    ys = hessian.YS[-1]
    yy = hessian.YY[-1]

    dot = vs.dot
    addmul= vs.addmul
    mul = vs.mul
    pow = vs.pow

    D0yy = dot(mul(y, D0), y)

    if pre_scaled: D0 = mul(ys / D0yy, D0)

    invD0 = pow(D0, -1)

    t = addmul(invD0, pow(y, 2), 1 / ys)
    t = addmul(t, pow(mul(s, invD0), 2), -1 / dot(mul(s, invD0), s))

    D1 = pow(t, -1)

    if post_scaled:
        D1yy = dot(mul(y, D1), y)
        D1 = mul(D1, ys / D1yy)

    return D1

def pre_scaled_direct_bfgs(vs, hessian):
    """ M1QN3.B2 """
    return direct_bfgs(vs, hessian, pre_scaled=True)

def post_scaled_direct_bfgs(vs, hessian):
    """ Modified from M1QN3.B2; Recommended by 
        VEERSE, AUROUX & FISHER, for fewer iterations.
    """
    return direct_bfgs(vs, hessian, post_scaled=True)

def inverse_dfp(vs, hessian, pre_scaled=False, post_scaled=False):
    """ 
        M1QN3.C and M1QN3.C2 in GL Equation 4.8, 4.10;
        and VAF post update scaling.
        This is not applicable since we do not implement DFP.
    """
    assert len(hessian.S) > 0

    D0 = hessian.D
    s = hessian.S[-1]
    y = hessian.Y[-1]
    ys = hessian.YS[-1]
    yy = hessian.YY[-1]

    dot = vs.dot
    mul = vs.mul
    addmul = vs.addmul
    pow = vs.pow

    yD0 = mul(y, D0)
    D0yy = dot(yD0, y)

    if pre_scaled: D0 = mul(ys / D0yy, D0)

    t = addmul(D0, pow(s, 2), 1 / ys)
    t = addmul(t,  pow(yD0, 2), 1/ D0yy)

    D1 = t

    if post_scaled:
        yD1 = mul(y, D1)
        D1yy = dot(yD1, y)
        D1 = mul(D1, ys / D1yy)

    return D1

def pre_scaled_inverse_dfp(vs, hessian):
    """ M1QN3.C2, we didn't implement DFP."""

    return inverse_dfp(vs, hessian, pre_scaled=True)

def post_scaled_inverse_dfp(vs, hessian):
    """ post-update version of M1QN3.C2.  we didn't implement DFP."""

    return inverse_dfp(vs, hessian, post_scaled=True)

class LBFGSHessian(object):
    def __init__(self, vs, m, diag_update=scalar, rescale_diag=False):
        """ D is a vector represents the initial diagonal. """
        self.m = m
        self.Y = [] # Delta G
        self.S = [] # Delta S
        self.YS = []
        self.YY = []
        self.D = 1.
        self.vs = vs
        self.diag_update = diag_update
        self.rescale_diag = rescale_diag

    def copy(self):
        r = LBFGSHessian(vs=self.vs, m=self.m, diag_update=self.diag_update, rescale_diag=self.rescale_diag)
        r.Y = self.Y[:]
        r.S = self.S[:]
        r.YS = self.YS[:]
        r.YY = self.YY[:]
        r.D = self.vs.copy(self.D)
        return r

    def hvp(self, v):
        """ Inverse of Hessian dot any vector; lowercase h indicates it is the inverse """
        q = v

        dot = self.vs.dot
        addmul = self.vs.addmul
        mul = self.vs.mul

        if len(self.Y) == 0: # first step
            return mul(v, self.D)

        alpha = list(range(len(self.Y)))
        beta = list(range(len(self.Y)))

        if self.YY[-1] == 0 or self.YS[-1] == 0: # failed LBFGS
            return None

        for i in range(len(self.Y) - 1, -1, -1):
            alpha[i] = dot(self.S[i], q) / self.YS[i]
            q = addmul(q, self.Y[i], -alpha[i])

        D = self.D

        if self.rescale_diag:
            Dyy = dot(mul(self.Y[-1], D), self.Y[-1])
            D = mul(self.YS[-1]/ Dyy, D)

        z = addmul(0, q, D)
        for i in range(len(self.Y)):
            beta[i] = 1.0 / self.YS[i] * dot(self.Y[i], z)
            z = addmul(z, self.S[i], (alpha[i] - beta[i]))

        return z

    def update(self, Px0, Px1, Pg0, Pg1):
        dot = self.vs.dot
        addmul = self.vs.addmul
        mul = self.vs.mul

        y = addmul(Pg1, Pg0, -1)
        s = addmul(Px1, Px0, -1)

        ys = dot(y, s)
        yy = dot(y, y)

        if yy == 0 or ys == 0:
            # refuse to add a degenerate mode.
            return

        self.Y.append(y)
        self.S.append(s)
        self.YS.append(ys)
        self.YY.append(yy)

        if len(self.Y) > self.m:
            del self.Y[0]
            del self.S[0]
            del self.YS[0]
            del self.YY[0]

        self.D = self.diag_update(self.vs, self)

    def __repr__(self):
        return "LBFGSHessian(len(Y)=%d, m=%d)" % (len(self.Y), self.m)

class LBFGSFailure(StopIteration):
    def __init__(self, message):
        self.message = message
        StopIteration.__init__(self)

class LBFGS(Optimizer):
    optimizer_defaults = {
        'maxiter' : 100000,
        'conviter' : 6,
        'm' : 6,
        'linesearch' : backtrace,
        'linesearchiter' : 100,
        'regulator' : simpleregulator,
        'diag_update' : post_scaled_direct_bfgs,
        'rescale_diag' : False,
    }

    def start(self, problem, state, x0):
        prop = Optimizer.start(self, problem, state, x0)
        prop.B = None
        prop.z = prop.Pg
        # carry over the gradient descent search radius
        prop.r1 = getattr(state, 'r1', 1.0)
        return prop

    def move(self, problem, state, prop):
        addmul = problem.vs.addmul
        dot = problem.vs.dot

        prop.complete(state)

        state.B = prop.B
        state.z = prop.z
        state.r1 = prop.r1

        if state.B is None:
            state.B = LBFGSHessian(problem.vs, self.m, self.diag_update, self.rescale_diag)
        else:
            state.B.update(state.Px, prop.Px, state.Pg, prop.Pg)

        Optimizer.move(self, problem, state, prop)

    def single_iteration(self, problem, state):
        """ Line by line translation of the LBFGS on wikipedia """
        addmul = problem.vs.addmul
        dot = problem.vs.dot
        mul = problem.vs.mul

        r1 = state.r1

        if state.Pgnorm == 0:
            raise RuntimeError("gnorm is zero. This shall not happen because terminated() checks for this")

        try:
            # use old LBFGSHessian, and update it
            B = state.B

            # no hessian approximation yet, use GD
            if len(B.Y) == 0:
                raise LBFGSFailure("no lbfgs")

            z = B.hvp(state.Pg)

            # hvp cannot be computed, recover
            if z is None:
                raise LBFGSFailure("hvp failed")

            znorm = dot(z, z) ** 0.5
            theta = dot(z, state.Pg) / (state.Pgnorm * znorm)
            if theta < 0.0:
                # purge the hessian approximation
                B = LBFGSHessian(problem.vs, self.m, self.diag_update, self.rescale_diag)
                raise LBFGSFailure("lbfgs misaligned theta = %g" % (theta,))

            # LBFGS should have been good, so we shall not search too many times.
            prop, r2 = self.linesearch(problem, state, z, 1.0, maxiter=3)

            # failed line search, recover
            if prop is None:
                B = LBFGSHessian(problem.vs, self.m, self.diag_update, self.rescale_diag)
                raise LBFGSFailure("lbfgs linesearch failed theta=%g " % (theta, ))

            # print('BFGS Starting step = %0.2e, step moved = %0.2e'%(r2))

            #if LBFGS is not moving, try GD
            if problem.check_convergence(state.y, prop.y):
                raise LBFGSFailure("lbfgs no improvement")
                # print('LBFGS Starting step = %0.2e, abort LBGFS at step = %0.2e'%(r2, r1))

        except LBFGSFailure as e:

            z = state.Pg

            r1max = self.regulator(problem, state, z)
            r1max = min(r1max, state.r1 * 2)

            prop, r1 = self.linesearch(problem, state, state.Pg, r1max, maxiter=self.linesearchiter)

            # failed GD
            if prop is None:
                r1 = state.r1
                return None

            prop.message = e.message
            # print('GD Starting step = %0.2e, step moved = %0.2e'%(r1max, r1))


        prop.B = B
        prop.z = z
        prop.r1 = r1

        return prop



import warnings

warnings.warn("vmad2 has been deprecated and moved to abopt.legacy.vmad2. "
              "Replace abopt.vmad2 with abopt.legacy.vmad2. "
              "Use the new python package vmad for new models.", FutureWarning)

from abopt.legacy.vmad2 import *



import numpy
import warnings

class VM(object):

    def copy_var(self, a):
        """ override for copying a variable """
        try:
            return a.copy()
        except:
            return 1.0 * a

    def iadd_var(self, a, b):
        """ override for inplace adding to a variable """
        a[...] += b

    @staticmethod
    def inspect(record):
        return str(list((k, "0x%0X" % id(v)) for k, v in record.items()))

    @staticmethod
    def microcode(fout=[], fin=[], lout=[]):
        """ Declare a subclass member function as a microcode.

            A microcode is a function with names of input and names of output
            It always takes a frontier argument and a list of *args.

            lout is the list of 'literal' outputs (do not participate in gradients)

            >>> @VM.microcode(fin=['x'], fout=['x'])
            >>> def mul(self, frontier, factor):
            >>>    frontier['x'] = factor * frontier['x']

            The backtrace gradient microcode shall be
            >>> @mul.grad
            >>> def gmul(self, frontier, factor):
            >>>    frontier['^x'] = factor * frontier['^x']

            Notice that `^` denotes the backtraced gradient against a variable.
            The same name can be both lvalue and rvalue
        """
        def decorator(func):
            def gradient_decorator(func1):
                func1.gout = ['^' + v for v in fin]
                func1.gin  = ['^' + v for v in fout]
                func1.fin = func.fin
                func1.fout = []
                func1.lout = []
                func.g = func1
                return func1

            func.fout = fout
            func.fin = fin
            func.lout = lout
            func.grad = gradient_decorator
            return func
        # no parameters, just a plain decorator
        if hasattr(fout, "__call__"):
            func = fout
            return decorator(func)
        return decorator

    def __init__(self):
        self._microcodes = []

    def __len__(self):
        return len(self._microcodes) + 1

    @property
    def microcodes(self):
        # always append a sentinal for final connection to the output.
        r = self._microcodes + [(None, ())]
        return r

    def push(self, op, v, args):
        """ Append to the microcode list of the VM

            Use this to build complicated microcode sequences.

            >>> vm.push('mul', 3.0)
            >>> vm.push('mul', 2.0)
        """
        for name, impl in self.__class__.__dict__.items():
            if name == op:
                self._microcodes.append((impl, args))
                break 
        else:
            raise AttributeError("code %s is not found" % op)

    def compute(self, fout, init, tape=None, monitor=None):
        """
            Run the list of microcodes with `init` dictionary as input
            Record the results on `tape` (list), and return a dictionary
            contains variables named in fout (list).
            The items in the init shall support [...] or the '+' operator

            >>> vm.compute(['x'], {'x': numpy.ones(10)})

        """

        frontier = {}
        frontier.update(init)
        literals = set()

        for step, (impl, args) in enumerate(self.microcodes):
            if impl is None:
                # sentinal,
                # add the final fout into the tape
                # to properly update refcount of the input gradients

                impl = lambda self, frontier: None
                # skip the literal from fin, even if it is requested from fout.
                impl.fin = [v for v in fout if v not in literals]
                impl.fout = []
                impl.lout = []

            for name in impl.fin:
                if name in literals:
                    raise RuntimeError(
                    "An input of microcode `%s`, `%s` is marked as literal by previous steps. The machine is likely buggy." % (name, impl.__name__))

            record = {}

            for var in set(impl.fin) | literals:
                record[var] = frontier[var]

            if monitor:
                monitor(step, len(self), impl, args, VM.inspect(record))

            if tape is not None:
                # replace the frontier with a copy -- such that the one on tape is preserved.
                for var in record:
                    if var in impl.fout:
                        # save a copy of the variables for in-place operations.
                        copy = self.copy_var(record[var])
                        assert id(copy) != id(record[var])
                        frontier[var] = copy

                tape.append(record)

            literals |= set(impl.lout)
            impl(self, frontier, *args)

        if tape is not None:
            # record fout literals
            # for future gradient
            tape.append(fout)
            tape.append(literals)

        r = {}
        for name in fout:
            r[name] = frontier[name]
        return r

    @staticmethod
    def _refcount(tape):
        """ count number of references to any variables on a tape """
        d = {}
        for record in tape:
            for name, value in record.items():
                # note that all variables on the record are alive,
                # in this case id uniquely determines the values.
                uid = id(value)
                d[uid] = d.get(uid, 0) + 1
        return d

    def gradient(self, gout, ginit, tape, monitor=None):
        """ Backtrace the gradient from ginit (dict).

            tape is obtained from a prevoius call to `compute`.

            Returns a dict contains the requested gradients in gout (list).

            >>> tape = []
            >>> vm.compute(['x'], {'x': numpy.ones(10)})
            >>> vm.gradient(['^x'], {'^x', numpy.ones(10)}, tape)
        """

        sentinal = [(None, ())]

        fout = tape[-2]
        literals = tape[-1]

        refcount = self._refcount(tape[:-2])

        #for i, n in refcount.items():
            #print ('0X%X' %i, n)

        # holding the partially computed gradient of variables
        partial = {}

        frontier = {}
        for (impl, args), record in reversed(list(zip(self.microcodes, tape))):
            d = {}
            d.update(record)
            d.update(frontier)

            if impl is None:
                # sentinal -- initialization
                # pump up d with the initial and be aware of missing items
                # in ginit -- if they are in gin we later set them to zero.

                impl = lambda : None
                impl.g = lambda self, d: None
                # literals are skipped from gradients, even if they are
                # requested by fout; if we keep them here
                # we will end up having warnings about unfinished partial
                # gradients. -- this is the only path literals gets mixed.
                # otherwise the some of the microcodes has inproperly decleared
                # literals in fin.
                impl.g.gin = ['^' + v for v in fout if v not in literals]
                impl.g.gout = impl.g.gin
                d.update(ginit)

            if monitor:
                monitor('gradient', impl, 'before', VM.inspect(record))

            
            for name in impl.g.gin:
                # non existing inputs are implied to start with gradients of VM.Zero
                if name not in d:
                    d[name] = VM.Zero

            impl.g(self, d, *args)

            # reduce the result
            for name in impl.g.gout:
                vname = name[1:]
                uid = id(record[vname])
                if monitor:
                    monitor('partial gradient', name, refcount[uid], d[name])
                pg = d[name]
                if uid not in partial or partial[uid] is VM.Zero:
                    # this is the first channel, create the gradient storage
                    partial[uid] = pg
                elif pg is not VM.Zero:
                    # this is an additional channel. cummulate the gradient.
                    self.iadd_var(partial[uid], pg)
                refcount[uid] = refcount[uid] - 1

                if refcount[uid] == 0:
                    # update the frontier with the new gradients
                    # we no longer need to save it on partial since cummulation is done.
                    frontier[name] = partial.pop(uid)
                    if monitor:
                        monitor('finalized gradient', name, frontier[name])

        for i, v in partial.items():
            warnings.warn('problem: remainging partial : 0X%X %d %s' % (i, refcount[i], str(v)), RuntimeWarning)

        r = {}
        for name in gout:
            r[name] = frontier[name]
        return r

#####################
#
# ZeroType.
#
# Zero * anything = Zero; Zero + anything = anything
#
# VM.Zero is used to suppress gradient computation
# backtracing from VM.Zero shall always give VM.Zero
# regardless of the microcode function; this may 
# get more complicated if a microcode function
# has multiple outputs -- e.g. if the backtrace gradient
# of some are not VM.Zero.
#
# In case it occurs in a regular Python expression,
# these operator overides ensure we can 
# propagate VM.Zero properly.
#
# In other cases the microcode functions may need
# to special case on variables that are VM.Zero.

def ZeroType():
    """ creates a special type of ZeroType; """
    def self(self, *args): return self
    def other(self, other): return other
    def zde(self, a): raise ZeroDivisionError
    def __sub__(self, a): return -a
    def __xor__(self, a): return ~a
    def __int__(self): return 0
    def __float__(self): return 0.0
    def __round__(self): return 0
    def __array__(self): return numpy.array(0)

    dict = {}
    for name, value in locals().items():
         if name.startswith("__"): dict[name] = value

    for name in [
        "neg", "pos", "abs", "invert", "complex",
        "mul", "rmul", "matmul", "rmatmul", 
        "mod", "divmod", "div", "truediv", "floordiv",
        "pow",
        "and", "rand", "lshift", "rlshift", "rshift", "rrshift",
        "getitem", "reversed"]:
        dict["__" + name + "__"] = self

    for name in [
        "rmod", "rdivmod", "rdiv", "rtruediv", "rfloordiv",
        "rpow", "rsub", "rxor"]:
        dict["__" + name + "__"] = zde

    for name in [
        "add", "radd", "or", "ror"]:
        dict["__" + name + "__"] = other

    return type("ZeroType", (object,), dict)

ZeroType = ZeroType()
Zero = ZeroType()

# now add Zero to VM for easier access.
VM.Zero = Zero



#!/usr/bin/env python
from distutils.core import setup

def find_version(path):
    import re
    # path shall be a plain ascii text file.
    s = open(path, 'rt').read()
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]",
                              s, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Version not found")

setup(name="abopt", version=find_version("abopt/version.py"),
      author="Grigor Aslanyan, Yu Feng, et al",
      maintainter="Yu Feng",
      maintainter_email="rainwoodman@gmail.com",
      description="Optimization of abstract data types in Python",
      zip_safe=True, # this should be pure python
      packages=["abopt",
                "abopt.linesearch",
                "abopt.algs",
                "abopt.algs.tests",
                "abopt.tests",
                "abopt.testing",
                "abopt.legacy",
                "abopt.legacy.tests",
               ],
      license='GPLv3',
      install_requires=['numpy', 
                        'scipy', # currently needed for 1d linesearch
                       ]
      )




from abopt.base import Proposal

# line search methods:
from .backtrace import backtrace
from .minpack import minpack
from .exact import exact

def simpleregulator(problem, state, z):
    # In LBFGS, the purpose of GD is to estimate the Hessian,
    # thus we do not want to move too far yet
    # limit it to 10 x of the original proposal
    dot = problem.vs.dot
    znorm = dot(z, z) ** 0.5

    rmax = 1.0

    if state.Pxnorm != 0:
        rmax = min(10 * state.Pxnorm / znorm, rmax)

    return rmax

def nullsearch(problem, state, z, rate, maxiter):
    """ A null line search that does not change the rate;

        Useful for e.g. stochastic gradient descent.

    """

    Px1 = addmul(state.Px, z, -rate)
    prop = Proposal(problem, Px=Px1, z=z).complete_y(state)
    return prob, rate



from __future__ import print_function
from numpy.testing import assert_allclose
from numpy.testing import assert_array_equal

from abopt.abopt2 import Problem, LBFGS

import pytest

try:
    import autograd
    from abopt.testing.autograd_problems import get_all_nd, get_all_2d
except ImportError:
    autograd = None
    def get_all_nd(): return []
    def get_all_2d(): return []

pytestmark = pytest.mark.skipif(autograd is None, reason="skipping test functions due to lack of autograd")

@pytest.mark.parametrize("case", get_all_2d())
def test_cases_fmin(case):
    case = case(nd=2)
    assert_allclose(case.function(case.xmin), case.fmin)

@pytest.mark.parametrize("case", get_all_2d())
def test_cases_grad_min(case):
    case = case(nd=2)
    grad = case.gradient(case.xmin)

    assert_allclose(grad, 0, atol=1e-4)

@pytest.mark.parametrize("case", get_all_2d())
def test_cases_grad_start(case):
    case = case(nd=2)
    grad = case.gradient(case.start)

    eps = 1e-5
    grad2 = []
    for i in range(len(case.start)):
        x = 1.0 * case.start
        x[i] += eps
        y2 = case.function(x)
        x[i] -= 2 * eps
        y1 = case.function(x)
        grad2.append((y2 - y1) / (2 * eps))

    assert_allclose(grad, grad2)

@pytest.mark.parametrize("case", get_all_2d())
def test_cases_domain(case):
    case = case(nd=2)
    grad = case.gradient(case.xmin)

    left, right = case.domain

    assert (case.xmin <= right).all()
    assert (case.xmin >= left).all()

    assert (case.xmin != case.start).all()

    assert (case.start <= right).all()
    assert (case.start >= left).all()

@pytest.mark.parametrize("case", get_all_nd())
def test_cases_lbfgs_nd(case):
    case = case(nd=1000)
    opt = LBFGS()

    x0 = case.start
    p = Problem(objective=case.function, gradient=case.gradient)

    r = opt.minimize(p, x0, monitor=print)

    assert r['nit'] < 20 # Probably hitting slow convergence. Bug in optimizer?
    assert_allclose(r['y'], case.fmin, atol=case.yatol)
    assert_allclose(r['x'], case.xmin, atol=case.xatol)

@pytest.mark.parametrize("case", get_all_2d())
def test_cases_lbfgs_2d(case):
    case = case(nd=2)
    opt = LBFGS()

    x0 = case.start
    p = Problem(objective=case.function, gradient=case.gradient)

    r = opt.minimize(p, x0, monitor=print)

    assert r['nit'] < 20 # Probably hitting slow convergence. Bug in optimizer?
    assert_allclose(r['y'], case.fmin, atol=case.yatol)
    assert_allclose(r['x'], case.xmin, atol=case.xatol)



from abopt.base import Proposal

def minpack(problem, state, z, rate, maxiter, c1=1e-4, c2=0.9, amax=50):
    """"
    Notes
    -----
    Uses the line search algorithm to enforce strong Wolfe
    conditions.  See Wright and Nocedal, 'Numerical Optimization',
    1999, pg. 59-60.

    For the zoom phase it uses an algorithm by [...].

    """
    vs = problem.vs
    addmul = vs.addmul
    dot = vs.dot

    def phi(alpha):
        state.fev = state.fev + 1
        Px1 = addmul(state.Px, z, -alpha)
        x1 = problem.Px2x(Px1)
        y1 = problem.f(x1)
        # print('phi', -alpha, y1, state.y)
        return y1

    Pgval = [(state.g, state.Pg), (state.x, state.Px)]

    def derphi(alpha):
        state.gev = state.gev + 1
        Px1 = addmul(state.Px, z, -alpha)
        x1 = problem.Px2x(Px1)
        g1 = problem.g(x1)
        Pg1 = problem.g2Pg(g1)
        Pgval[0] = (g1, Pg1)
        Pgval[1] = (x1, Px1)
    #    print('derphi', x1, g1)
        return -dot(Pg1, z)

    derphi0 = -dot(state.Pg, z)

    y0 = state.y_[-1]
    if len(state.y_) > 1:
        y0_prev = state.y_[-2]
    else:
        y0_prev = None

    alpha_star, phi_star, old_fval, derphi_star = scalar_search_wolfe2(
            phi, derphi, y0, y0_prev, derphi0, c1, c2, amax * rate)

    if derphi_star is None:
        #raise ValueError('The line search algorithm did not converge')
        return None, None
    else:
        # derphi_star is a number (derphi) -- so use the most recently
        # calculated gradient used in computing it derphi = gfk*pk
        # this is the gradient at the next step no need to compute it
        # again in the outer loop.
        (g1, Pg1), (x1, Px1) = Pgval
        prop = Proposal(problem, g=g1, Pg=Pg1, x=x1, Px=Px1, y=phi_star, z=z)
        return prop, alpha_star

def scalar_search_wolfe2(phi, derphi=None, phi0=None,
                         old_phi0=None, derphi0=None,
                         c1=1e-4, c2=0.9, amax=50, maxiter=10):
    """Find alpha that satisfies strong Wolfe conditions.

    alpha > 0 is assumed to be a descent direction.

    Parameters
    ----------
    phi : callable f(x)
        Objective scalar function.
    derphi : callable f'(x), optional
        Objective function derivative (can be None)
    phi0 : float, optional
        Value of phi at s=0
    old_phi0 : float, optional
        Value of phi at previous point
    derphi0 : float, optional
        Value of derphi at s=0
    c1 : float, optional
        Parameter for Armijo condition rule.
    c2 : float, optional
        Parameter for curvature condition rule.
    amax : float, optional
        Maximum step size

    Returns
    -------
    alpha_star : float or None
        Best alpha, or None if the line search algorithm did not converge.
    phi_star : float
        phi at alpha_star
    phi0 : float
        phi at 0
    derphi_star : float or None
        derphi at alpha_star, or None if the line search algorithm
        did not converge.

    Notes
    -----
    Uses the line search algorithm to enforce strong Wolfe
    conditions.  See Wright and Nocedal, 'Numerical Optimization',
    1999, pg. 59-60.

    For the zoom phase it uses an algorithm by [...].

    """

    if phi0 is None:
        phi0 = phi(0.)

    if derphi0 is None and derphi is not None:
        derphi0 = derphi(0.)

    alpha0 = 0
    if old_phi0 is not None and derphi0 != 0:
        alpha1 = min(1.0, 1.01*2*(phi0 - old_phi0)/derphi0)
    else:
        alpha1 = 1.0

    if alpha1 < 0:
        alpha1 = 1.0

    if alpha1 == 0:
        # This shouldn't happen. Perhaps the increment has slipped below
        # machine precision?  For now, set the return variables skip the
        # useless while loop, and raise warnflag=2 due to possible imprecision.
        alpha_star = None
        phi_star = phi0
        phi0 = old_phi0
        derphi_star = None

    phi_a1 = phi(alpha1)
    #derphi_a1 = derphi(alpha1)  evaluated below

    phi_a0 = phi0
    derphi_a0 = derphi0

    i = 0
    while alpha0 < amax:
        #print('alpha0', alpha0)
        if alpha1 == 0:
            break
        if (phi_a1 > phi0 + c1 * alpha1 * derphi0) or \
           ((phi_a1 >= phi_a0) and (i > 1)):
            alpha_star, phi_star, derphi_star = \
                        _zoom(alpha0, alpha1, phi_a0,
                              phi_a1, derphi_a0, phi, derphi,
                              phi0, derphi0, c1, c2, maxiter)
            break

        derphi_a1 = derphi(alpha1)
        if (abs(derphi_a1) <= -c2*derphi0):
            alpha_star = alpha1
            phi_star = phi_a1
            derphi_star = derphi_a1
            break

        if (derphi_a1 >= 0):
            alpha_star, phi_star, derphi_star = \
                        _zoom(alpha1, alpha0, phi_a1,
                              phi_a0, derphi_a1, phi, derphi,
                              phi0, derphi0, c1, c2, maxiter)
            break

        alpha2 = 2 * alpha1   # increase by factor of two on each iteration
        alpha0 = alpha1
        alpha1 = alpha2
        phi_a0 = phi_a1
        phi_a1 = phi(alpha1)
        derphi_a0 = derphi_a1
        i = i + 1
    else:
        # stopping test maxiter reached
        alpha_star = alpha1
        phi_star = phi_a1
        derphi_star = None
        warn('The line search algorithm did not converge', LineSearchWarning)

    return alpha_star, phi_star, phi0, derphi_star

from scipy.optimize.linesearch import _cubicmin, _quadmin

def _zoom(a_lo, a_hi, phi_lo, phi_hi, derphi_lo,
          phi, derphi, phi0, derphi0, c1, c2, maxiter):
    """
    Part of the optimization algorithm in `scalar_search_wolfe2`.
    """

    #print('_zoom', a_lo, a_hi, phi_lo, phi_hi)
    i = 0
    delta1 = 0.2  # cubic interpolant check
    delta2 = 0.1  # quadratic interpolant check
    phi_rec = phi0
    a_rec = 0
    while True:
        # interpolate to find a trial step length between a_lo and
        # a_hi Need to choose interpolation here.  Use cubic
        # interpolation and then if the result is within delta *
        # dalpha or outside of the interval bounded by a_lo or a_hi
        # then use quadratic interpolation, if the result is still too
        # close, then use bisection

        dalpha = a_hi - a_lo
        if dalpha < 0:
            a, b = a_hi, a_lo
        else:
            a, b = a_lo, a_hi

        # minimizer of cubic interpolant
        # (uses phi_lo, derphi_lo, phi_hi, and the most recent value of phi)
        #
        # if the result is too close to the end points (or out of the
        # interval) then use quadratic interpolation with phi_lo,
        # derphi_lo and phi_hi if the result is stil too close to the
        # end points (or out of the interval) then use bisection

        if (i > 0):
            cchk = delta1 * dalpha
            a_j = _cubicmin(a_lo, phi_lo, derphi_lo, a_hi, phi_hi,
                            a_rec, phi_rec)
        if (i == 0) or (a_j is None) or (a_j > b - cchk) or (a_j < a + cchk):
            qchk = delta2 * dalpha
            a_j = _quadmin(a_lo, phi_lo, derphi_lo, a_hi, phi_hi)
            if (a_j is None) or (a_j > b-qchk) or (a_j < a+qchk):
                a_j = a_lo + 0.5*dalpha

        # Check new value of a_j

        phi_aj = phi(a_j)
        if (phi_aj > phi0 + c1*a_j*derphi0) or (phi_aj >= phi_lo):
            phi_rec = phi_hi
            a_rec = a_hi
            a_hi = a_j
            phi_hi = phi_aj
        else:
            derphi_aj = derphi(a_j)
            if abs(derphi_aj) <= -c2*derphi0:
                a_star = a_j
                val_star = phi_aj
                valprime_star = derphi_aj
                break
            if derphi_aj*(a_hi - a_lo) >= 0:
                phi_rec = phi_hi
                a_rec = a_hi
                a_hi = a_lo
                phi_hi = phi_lo
            else:
                phi_rec = phi_lo
                a_rec = a_lo
            a_lo = a_j
            phi_lo = phi_aj
            derphi_lo = derphi_aj
        #print(a_j, phi_aj, a_lo, a_hi, phi0)
        i += 1
        if (i > maxiter):
            # Failed to find a conforming step size
            a_star = a_j
            val_star = phi_aj
            valprime_star = derphi(a_j)
            break
    return a_star, val_star, valprime_star
